
#include "ccInteractor.h"

void MainWindow::init(void) {
    interactor = NULL;
    QVBox *mainBox = new QVBox(this);
    mainBox->setGeometry(10,10,330,400);
  
    QPushButton *buttonOK = new QPushButton("Open 3Ds-File...", mainBox, "button_open");
    buttonOK->setGeometry(10,10,310,20);
    QHBox *h1box = new QHBox(mainBox);
    h1box->setGeometry(10,30,310,20);
    lsubdiv = new QLabel("&Number of Subdivisions:",h1box);
    noOfSubdivisions = new QSpinBox(h1box,"spin1");
    noOfSubdivisions->setMinValue(1);
    noOfSubdivisions->setMaxValue(20);
    lsubdiv->setBuddy(noOfSubdivisions);
    QButtonGroup *selectStartType = new QButtonGroup("Select initial mesh type", mainBox);
    selectStartType->setGeometry(10,60,310,80);
    QButtonGroup *selectSubdivMeth = new QButtonGroup("Select initial subdivision method", mainBox);
    selectSubdivMeth->setGeometry(10,110,310,140);
    triangle = new QRadioButton("Triangle", selectStartType);
    triangle->setGeometry(10, 20, 100, 20);
    triangle->setChecked(true);
    quad = new QRadioButton("Quad", selectStartType);
    quad->setGeometry(10, 50, 100, 20);
    quad->setChecked(false);
    selectStartType->insert(triangle, 0);
    selectStartType->insert(quad, 1);
    linear = new QRadioButton("Linear", selectSubdivMeth);
    linear->setGeometry(10,20,200,20);
    linear->setChecked(true);
    loop = new QRadioButton("Loop", selectSubdivMeth);
    loop->setGeometry(10,50,200,20);
    loop->setChecked(false);
    butterfly = new QRadioButton("Butterfly",selectSubdivMeth);
    butterfly->setGeometry(10,80,200,20);
    butterfly->setChecked(false);
    cc = new QRadioButton("CatmullClark",selectSubdivMeth);
    cc->setGeometry(10,110,200,20);
    cc->setChecked(false);

    selectSubdivMeth->insert(linear,0);
    selectSubdivMeth->insert(loop,1);
    selectSubdivMeth->insert(butterfly,2);
    selectSubdivMeth->insert(cc,3);
    QPushButton *buttonSubdivide = new QPushButton("Subdivision", mainBox, "button_subdivide");
    QPushButton *buttonSave = new QPushButton("Save Image",mainBox,"button_save");
    connect( buttonOK, SIGNAL(clicked()), this, SLOT(openFile()));
    connect( buttonSubdivide, SIGNAL(clicked()), this, SLOT(newSubdivision()));
    connect( buttonSave, SIGNAL(clicked()),this,SLOT(saveFile()));
}


void MainWindow::openFile(void) {

    QFileDialog *fd = new QFileDialog( this, "Open File", TRUE );
        
    fd->setMode( QFileDialog::ExistingFile );
    fd->setViewMode( QFileDialog::List );
    fd->setFilter( "Data File (*.3ds)" );

    if( fd->exec() == QDialog::Accepted ) {
        fileName = fd->selectedFile();
    }
}

void MainWindow::saveFile(void) {
    if(interactor != NULL) {
        SaveFile = QFileDialog::getSaveFileName("/home","Images (*.bmp *.xbm *.xpm *.png *.jpg *.jpeg )",
                                                this, "save file dialog", 
                                                "Choose a file name to save under");
        QImage *newImg = new QImage();
	*newImg = interactor->grabFrameBuffer();

        if ( SaveFile != NULL ) {
            if(SaveFile.endsWith("jpg") || 
               SaveFile.endsWith("JPG") || 
               SaveFile.endsWith("JPEG")||
               SaveFile.endsWith("jpeg")) { 
                newImg->save(SaveFile,"JPEG",100);
            } else {
                if(SaveFile.endsWith("bmp") ||
                   SaveFile.endsWith("BMP")) {
                    newImg->save(SaveFile,"BMP",100);
                } else {
                    if(SaveFile.endsWith("xbm") ||
                       SaveFile.endsWith("XBM")) {
                        newImg->save(SaveFile,"XBM",100);
                    } else {
                        if(SaveFile.endsWith("xpm") ||
                           SaveFile.endsWith("XPM")) {
                            newImg->save(SaveFile,"XPM",100);
                        } else {
                            newImg->save(SaveFile,"PNG",100);
                        }
                    }
                }
            }
        }
	delete newImg;
    }
}

void MainWindow::newSubdivision(void) {

    printf("start of new subdivision\n");

    importer = vtk3DSOurImporter::New();
    if (triangle->isChecked()) importer->SetPolys("triangle"); else importer->SetPolys("quad");
    importer->ComputeNormalsOn();
    importer->SetFileName(fileName);
    importer->Read();

    printf("Test\n");


    subdiv = (noOfSubdivisions->value());

    filter1 = NULL;
    filter2 = NULL;

    // create Render Window Interactor
    interactor = new RenderWindowInteractor(this);
    CHECK_PTR( interactor );
    interactor->resize( 400, 400 );
    interactor->setGeometry( 360, 10, 400, 400 );

    if(linear->isChecked())
        (vtkLinearSubdivisionFilter*)filter1 = vtkLinearSubdivisionFilter::New();
    if(loop->isChecked())
        (vtkLoopSubdivisionFilter*)filter2 = vtkLoopSubdivisionFilter::New();
    if(butterfly->isChecked())
        (vtkButterflySubdivisionFilter*)filter1 = vtkButterflySubdivisionFilter::New();
    if(cc->isChecked())
        (vtkCatmullClarkFilter*)filter2 = vtkCatmullClarkFilter::New();
  
    if((filter1 != NULL)&&(filter2 == NULL))
        filter1->SetNumberOfSubdivisions(subdiv);
    else if((filter2 != NULL)&&(filter1 == NULL))
        filter2->SetNumberOfSubdivisions(subdiv);
    else exit(1);
  
    vtkRenderer *renderer = vtkRenderer::New();
    vtkRenderWindow *renderWindow = vtkRenderWindow::New();
    renderWindow->AddRenderer(renderer);
    renderer->SetBackground(0,0,0); 

    vtkActorCollection *ac = importer->GetRenderer()->GetActors();
    ac->InitTraversal();
    vtkActor *dsActor=ac->GetNextActor();
    while (dsActor!=NULL) {
        vtkPolyData *polyData = (vtkPolyData*)(dsActor->GetMapper()->GetInputAsDataSet());

        vtkPolyDataMapper *figureMapper = vtkPolyDataMapper::New();
  
        if(filter1 != NULL) {
            filter1->SetInput(polyData);
            figureMapper->SetInput(filter1->GetOutput());
        }
        if(filter2 != NULL) {
            filter2->SetInput(polyData);
            figureMapper->SetInput(filter2->GetOutput());
        }
        
        vtkActor *actor = vtkActor::New();
    
        actor->SetMapper(figureMapper);
        actor->SetProperty(dsActor->GetProperty()); 

        // ========================================================
        // Here the Actors are added to our interactor
        // ========================================================
        interactor->GetRenderer()->AddActor(actor);

        dsActor=ac->GetNextActor();
    }

    interactor->GetRenderer()->ResetCamera();
    printf("show interactor\n");
    interactor->show();
    
}

int main( int argc, char* argv[] ) {

    // ========================================================
    // Set up the QT-Application and init QT GUI- and VTK-stuff
    // ========================================================
    printf("Application starting\n");
    QApplication myApp( argc, argv );
  
    // create Main Window
    MainWindow *mainWindow = new MainWindow();
    mainWindow->setGeometry(100, 50, 800, 600);
    mainWindow->setCaption("MMCG - Subdivision Surfaces");
  
    // generate the interaction-gui
    mainWindow->init();
  
    // declare main window and show it
    myApp.setMainWidget( mainWindow );
    printf("show main window\n");
    mainWindow->show();
  
    //mainWindow->openFile();
  
    //mainWindow->newSubdivision();
  
    // qt main event loop
    //printf("execute application\n");
    int nReturn = myApp.exec();
  
    // clean vtk stuff
    // ToDo
  
    return nReturn;
}
