#include <vtkConeSource.h>
#include <vtkPolyDataMapper.h>
#include <vtkRenderWindow.h>
#include <vtkLoopSubdivisionFilter.h>
#include <vtkTriangleFilter.h>
#include <vtkActorCollection.h>
#include <vtkPolyData.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkInterpolatingSubdivisionFilter.h>
#include <vtkLinearSubdivisionFilter.h>
#include <vtkButterflySubdivisionFilter.h>
#include <vtkActor.h>

#include <qmainwindow.h>
#include <qvbox.h>
#include <qmenubar.h>
#include <qaction.h>
#include <qfiledialog.h>
#include <qpushbutton.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qimage.h>
#include <qradiobutton.h>
#include <qtabbar.h>
#include <qstatusbar.h>
#include <qwidget.h>
#include <qlabel.h>
#include <qprogressbar.h>
#include <qslider.h>
#include <qbuttongroup.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qcolor.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qapplication.h>

#include "vtk3DSOurImporter.h"
#include "vtkCatmullClarkFilter.h"

#include "renderwindowinteractor.h"

/// creates the GUI with the help of the vtkandqt library

class MainWindow : public QMainWindow
{
    Q_OBJECT
private:
/// provides GUI and OpenGL functionality
    RenderWindowInteractor *interactor; 
/// Number of Subdivisions
    int subdiv; 
/// filename of *.3DS to open
    QString fileName; 
/// filename of image file to save
    QString SaveFile; 
/// indentifier for linear and butterfly subdivision method 
    vtkInterpolatingSubdivisionFilter *filter1; 
/// indentifier for catmull-clark and loop subdivision method 
    vtkApproximatingSubdivisionFilter *filter2; 
/// imports *.3DS files in vtk
    vtk3DSOurImporter *importer; 

    //-- GUI Elements -------------------
/// GUI element for selection of the number of subdivisions
    QSpinBox *noOfSubdivisions; 
/// GUI element: Radiobutton for selection of mesh type
    QRadioButton *triangle; 
/// GUI element: Radiobutton for selection of mesh type
    QRadioButton *quad; 
/// GUI element: Radiobutton for selection of sudivision type
    QRadioButton *loop; 
/// GUI element: Radiobutton for selection of subdivision type
    QRadioButton *linear; 
/// GUI element: Radiobutton for selection of subdivision type
    QRadioButton *butterfly; 
/// GUI element: Radiobutton for selection of subdivision type
    QRadioButton *cc; 
/// GUI element: label
    QLabel *lsubdiv; 

public:
    /// initializes GUI  
    void init(void); 
public slots:
/// Method for handling a call to open a file
  void openFile();
/// Method for handling a call to save an image 
  void saveFile();
/// Method for handling subdivision calls
  void newSubdivision(void); 
};
